from kognic.io.model.calibration.camera.common import BaseCameraCalibration, DistortionCoefficients
from kognic.io.model.calibration.common import CalibrationType


class FisheyeCalibration(BaseCameraCalibration):
    calibration_type = CalibrationType.FISHEYE.value
    distortion_coefficients: DistortionCoefficients
    xi: float



from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.model.calibration import Position, RotationQuaternion


class EgoVehiclePose(BaseSerializer):
    """Both `position` and `rotation` are with respect to the local coordinate system (LCS)."""
    position: Position
    rotation: RotationQuaternion



import logging

_setup = False
root_logger = None
kognic_logger = None

TEXT_FORMAT = '%(asctime)-15s %(levelname)-s [%(name)-25s] %(message)s'


def setup_logging(level: str):
    global _setup
    if _setup:
        raise RuntimeError("Logging is already setup")

    module_log_levels = dict()
    if level is not None:
        # force override
        module_log_levels["root"] = level
        module_log_levels["kognic"] = level

    global root_logger, kognic_logger
    root_logger = logging.getLogger()
    kognic_logger = root_logger.getChild("kognic")

    for (lg, lvl) in module_log_levels.items():
        logger = root_logger if lg == "root" else logging.getLogger(lg)
        logger.setLevel(logging.getLevelName(lvl))
    logging.basicConfig(format=TEXT_FORMAT)

    _setup = True



from typing import List, Optional

import kognic.io.model.projects as ProjectModel
from kognic.io.resources.abstract import IOResource


class ProjectResource(IOResource):
    """
    Project related information
    """

    def get_projects(self) -> List[ProjectModel.Project]:
        """
        Returns all projects connected to the users organization.

        :return List: List containing all projects connected to the user
        """
        json_resp = self._client.get("v1/projects")
        return [ProjectModel.Project.from_json(js) for js in json_resp]

    def get_project_batches(self, project: str) -> List[ProjectModel.ProjectBatch]:
        """
        Returns all `batches` for the `project`.

        :return List: List containing all batches
        """
        json_resp = self._client.get(f"v1/projects/{project}/batches")
        return [ProjectModel.ProjectBatch.from_json(js) for js in json_resp]

    def create_batch(self, project: str, batch: str, publish_previous_batches: bool = False) -> ProjectModel.ProjectBatch:
        """
        Create new `batch` for the `project`. Publishes the previous batch if it is still open.

        :param project Project identifier
        :param batch Batch identifier
        :param publish_previous_batches set to true to publish previous batches, setting them to a Ready status
        :return ProjectBatch: The newly created batch
        """
        json_resp = self._client.post(f"v1/projects/{project}/batches/{batch}", json={"publishPreviousBatches": publish_previous_batches})
        return ProjectModel.ProjectBatch.from_json(json_resp)

    def publish_batch(self, project: str, batch: str) -> ProjectModel.ProjectBatch:
        """
        Publish input batch, marking the input batch ready for annotation.
        After publishing, no more inputs can be added to the input batch

        :param project Project identifier
        :param batch Batch identifier
        :return ProjectBatch: Updated input batch
        """
        json_resp = self._client.post(f"v1/projects/{project}/batches/{batch}/publish")
        return ProjectModel.ProjectBatch.from_json(json_resp)

    def get_annotation_types(self, project: str, batch: Optional[str] = None) -> List[str]:
        """
        Returns all available annotation types for the project.
        If batch included all available annotation types for the batch will be returned.
        """
        if batch:
            json_resp = self._client.get(f"v1/projects/{project}/batches/{batch}/annotation-types")
        else:
            json_resp = self._client.get(f"v1/projects/{project}/annotation-types")
        return [anno for anno in json_resp]



from datetime import datetime
from typing import Dict, Mapping, Optional

from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.model.calibration.camera.fisheye_calibration import FisheyeCalibration
from kognic.io.model.calibration.camera.kannala_calibration import KannalaCalibration
from kognic.io.model.calibration.camera.pinhole_calibration import PinholeCalibration
from kognic.io.model.calibration.camera.principal_point_distortion_calibration import \
    PrincipalPointDistortionCalibration
from kognic.io.model.calibration.common import BaseCalibration, CalibrationType
from kognic.io.model.calibration.lidar.lidar_calibration import LidarCalibration
from kognic.io.util import ts_to_dt

calibration_factory = {
    CalibrationType.LIDAR: LidarCalibration,
    CalibrationType.PINHOLE: PinholeCalibration,
    CalibrationType.KANNALA: KannalaCalibration,
    CalibrationType.FISHEYE: FisheyeCalibration,
    CalibrationType.PRINCIPALPOINTDIST: PrincipalPointDistortionCalibration
}


class SensorCalibration(BaseSerializer):
    external_id: str
    calibration: Dict[str, BaseCalibration]

    def to_dict(self):
        return {'externalId': self.external_id, 'calibration': {k: v.to_dict(by_alias=False) for (k, v) in self.calibration.items()}}


class SensorCalibrationEntry(BaseSerializer):
    id: str
    external_id: str
    created: datetime
    calibration: Optional[Mapping[str, BaseCalibration]]

    @classmethod
    def from_json(cls, js: dict):
        calibrations = js.get("calibration", {})

        calibration = {}
        for sensor, calib in calibrations.items():
            calibration[sensor] = cls._parse_calibration(calib)
        return SensorCalibrationEntry(id=js["id"], external_id=js["externalId"], created=ts_to_dt(js["created"]), calibration=calibration)

    @staticmethod
    def _parse_calibration(calibration: Dict) -> BaseCalibration:

        calibration_type = calibration.get("calibration_type")
        CalibrationModel = calibration_factory.get(calibration_type)
        if CalibrationModel is None:
            raise TypeError(f"Unable to parse calibration type: {calibration_type}")

        return CalibrationModel.parse_obj(calibration)



import io
import json
import logging
from typing import Dict, List, Optional, Tuple, Mapping

from kognic.base_clients.cloud_storage import FileResourceClient
from kognic.base_clients.cloud_storage import UploadSpec
from kognic.base_clients.http_client import HttpClient

import kognic.io.model as IOM
from kognic.io.model.ego import IMUData
from kognic.io.model.ego.validated_imu_data import ValidatedIMUData, ValidateIMUDataRequest
from kognic.io.model.input.feature_flags import FeatureFlags
from kognic.io.model.input.initialized_input import InitializedInput
from kognic.io.model.input.input_from_scene_request import InputFromSceneRequest
from kognic.io.model.input.resources.resource import Resource

log = logging.getLogger(__name__)

INPUTS_ROUTE = "v1/inputs"


class CreateableIOResource:

    def __init__(self, client: HttpClient, file_resource_client: FileResourceClient):
        super().__init__()
        self._client = client
        self._file_resource_client = file_resource_client

    def _post_input_request(
            self,
            resource_path: str,
            body: dict,
            project: Optional[str],
            batch: Optional[str],
            annotation_types: Optional[List[str]] = None,
            imu_data: Optional[List[IMUData]] = None,
            resources: Optional[Mapping[str, Resource]] = None,
            dryrun: bool = False,
            feature_flags: Optional[FeatureFlags] = None
    ) -> Optional[IOM.InputJobCreated]:
        """
        Send input to Input API. if dryrun is true, only validation is performed
        Otherwise, returns `InputJobCreated`
        """

        body, initialized_input = self._get_scene_uuid(body)

        body, validated_imu_data = self._validate_imu_data(body, imu_data, initialized_input.scene_uuid, dryrun)

        body = self._append_annotation_types(body, annotation_types)

        feature_flags = feature_flags if feature_flags is not None else FeatureFlags.defaults()

        body = self._append_feature_flags(body, feature_flags)

        input_job_json = self._validate_and_create_scene(
            resource_path=resource_path, body=body, project=project, batch=batch, dryrun=dryrun
        )

        if not dryrun:

            response = IOM.InputJobCreated.from_json(input_job_json)

            local_files_uploaded = self._upload_local_files(resources, response)

            local_files_uploaded += self._upload_imu_data(validated_imu_data)

            if local_files_uploaded:
                self._client.post(f"{INPUTS_ROUTE}/{response.input_uuid}/actions/commit", discard_response=True)

            return response

    def _get_scene_uuid(self, body: Dict) -> Tuple[Dict, InitializedInput]:
        json_resp = self._client.post(INPUTS_ROUTE + "/initialize")

        init_input = InitializedInput.from_json(json_resp)
        body["internalId"] = init_input.scene_uuid

        return body, init_input

    def _validate_imu_data(self, body: Dict, imu_data: Optional[List[IMUData]], input_uuid: str,
                           dryrun: bool) -> Tuple[Dict, Optional[ValidatedIMUData]]:
        if imu_data is None or len(imu_data) == 0:
            return body, None

        imu_request = ValidateIMUDataRequest(imu_data=imu_data, internal_id=input_uuid)
        json_resp = self._client.post(f"v1/imu", json=imu_request.to_dict(), dryrun=dryrun)

        valid_imu_data = None
        if json_resp is not None and not dryrun:
            json_resp["imuData"] = imu_data
            valid_imu_data = ValidatedIMUData.from_json(json_resp)
            body["imuData"] = valid_imu_data.resource_id

        return body, valid_imu_data

    def _append_annotation_types(self, body: Dict, annotation_types: Optional[List[str]]) -> Dict:
        if annotation_types:
            body['annotationTypes'] = annotation_types
        return body

    def _append_feature_flags(self, body: Dict, feature_flags: Optional[FeatureFlags]) -> Dict:
        if feature_flags:
            body['featureFlags'] = feature_flags.to_dict()
        return body

    def _validate_and_create_scene(
            self, *, resource_path: str, body: dict, project: Optional[str], batch: Optional[str], dryrun: bool = False
    ) -> Optional[Dict]:

        log.debug("POST:ing to %s input %s", resource_path, body)
        request_url = self._resolve_request_url(resource_path, project, batch)
        return self._client.post(request_url, json=body, dryrun=dryrun)

    def _upload_local_files(self, resources: Optional[Mapping[str, Resource]], response: IOM.InputJobCreated) -> bool:
        num_signed = len(response.files)
        num_requested = len(resources) if resources is not None else 0
        if num_signed > 0:
            assert num_signed == num_requested, f"File count discrepancy: {num_requested} != {num_signed}"
            uploads = {}
            for resource_id, signed_url in response.files.items():
                uploads[resource_id] = CreateableIOResource.create_upload_spec(signed_url, resources[resource_id])
            self._file_resource_client.upload_files(uploads)
            return True
        return False

    def _upload_imu_data(self, validated_imu_data: Optional[ValidatedIMUData]) -> bool:
        if validated_imu_data is not None and validated_imu_data.signed_url is not None:
            file = io.BytesIO(json.dumps(validated_imu_data.serialize_imu_data()).encode("utf-8"))
            self._file_resource_client.upload_json(file, validated_imu_data.signed_url)
            return True
        return False

    @staticmethod
    def _resolve_request_url(resource_path: str, project: Optional[str] = None, batch: Optional[str] = None) -> str:
        """
        Resolves which request url to use for input based on if project and batch is specified
        """
        url = INPUTS_ROUTE + "/"

        if project is not None:
            url += f"project/{project}/"
            if batch is not None:
                url += f"batch/{batch}/"

        url += resource_path

        return url

    @staticmethod
    def create_upload_spec(destination: str, res: Resource):
        return UploadSpec(destination=destination, content_type=res.content_type, filename=res.client_filename,
                          data=res.file_data.data if res.file_data is not None else None,
                          callback=res.file_data.callback if res.file_data is not None else None)

    def create_from_scene(
            self,
            scene_uuid: str,
            project: str,
            batch: Optional[str] = None,
            annotation_types: Optional[List[str]] = None,
            dryrun: bool = False
    ):
        """
        Create inputs from a scene. Note that, if a pre-annotation has been added for the scene, it will be included for
        all inputs.

        :param scene_uuid: uuid of the scene to create inputs from
        :param project: the project to add the inputs to
        :param batch: the batch to add the inputs to. Will default to the latest open batch if 'None'.
        :param annotation_types: list of annotation types to add inputs for. Will default to all annotation types for
            the project and batch if 'None' or empty
        :param dryrun: whether to do a dry-run or not
        """
        if annotation_types is None:
            annotation_types = []
        create_request = InputFromSceneRequest(scene_uuid=scene_uuid, annotation_types=annotation_types,
                                               project=project, batch=batch)
        return self._client.post(INPUTS_ROUTE, json=create_request.to_dict(), dryrun=dryrun, discard_response=True)



from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.model.calibration.camera.common import BaseCameraCalibration
from kognic.io.model.calibration.common import CalibrationType


class PrincipalPointDistortionCoefficients(BaseSerializer):
    k1: float
    k2: float


class PrincipalPoint(BaseSerializer):
    x: float
    y: float


class DistortionCenter(BaseSerializer):
    x: float
    y: float


class PrincipalPointDistortionCalibration(BaseCameraCalibration):
    calibration_type = CalibrationType.PRINCIPALPOINTDIST.value
    principal_point_distortion_coefficients: PrincipalPointDistortionCoefficients
    distortion_center: DistortionCenter
    principal_point: PrincipalPoint



from datetime import datetime
from typing import Dict

from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.util import ts_to_dt


class Annotation(BaseSerializer):
    input_uuid: str
    annotation_type: str
    created: datetime
    content: Dict

class PartialAnnotation(BaseSerializer):
    input_uuid: str
    annotation_type: str
    created: datetime
    uri: str

    def to_annotation(self, content: Dict) -> Annotation:
        return Annotation(input_uuid=self.input_uuid, annotation_type=self.annotation_type, created=self.created,
                          content=content)



from typing import List, Optional, Mapping

from kognic.io.model import UnixTimestampNs
from kognic.io.model.input.abstract.base_frame import BaseFrame
from kognic.io.model.input.resources.point_cloud import PointCloud


class Frame(BaseFrame):
    point_clouds: List[PointCloud]
    unix_timestamp: Optional[UnixTimestampNs] = None

    @property
    def resources(self) -> Mapping[str, PointCloud]:
        return {p.resource_id: p for p in self.point_clouds}



from typing import List, Optional

from kognic.io.model.base_serializer import BaseSerializer


class InputFromSceneRequest(BaseSerializer):
    scene_uuid: str
    annotation_types: List[str]
    project: str
    batch: Optional[str]



from typing import Mapping, Optional, Union

from pydantic import BaseModel

MetaDataMapping = Mapping[str, Union[int, float, str, bool]]


class MetaData(BaseModel):
    """
    Container for metadata belonging to the input.
    The attributes of this class are reserved keywords
    with assiciated functionality in the Kognic platform.

    Attributes:
        region: A string indicating the region the data was collected in. 
            If there are annotation instructions associated with a specific
            region, talk to your contact at Kognic to sync what should be
            specified in the region attribute.
    """
    region: Optional[str]

    class Config:
        extra = "allow"


AllowedMetaData = Union[MetaDataMapping, MetaData]


def metadata_to_dict(metadata: AllowedMetaData) -> MetaDataMapping:
    if isinstance(metadata, MetaData):
        return metadata.dict()
    return metadata



from pydantic import validator

from kognic.io.model.input.resources.resource import Resource
from kognic.io.resources.input.file_data import FileData

lidar_sensor_default = "lidar"


class PointCloud(Resource):
    sensor_name: str = lidar_sensor_default

    @validator('file_data', pre=True)
    def format_check(cls, value: FileData):
        if not value.format.is_pointcloud:
            raise ValueError(f"Invalid format for pointcloud data: {value.format}")
        return value



from typing import List, Mapping

from kognic.io.model import PointCloud
from kognic.io.model.input.abstract import BaseInputWithIMUData
from kognic.io.model.input.lidars_sequence.frame import Frame
from kognic.io.model.input.metadata.metadata import AllowedMetaData


class LidarsSequence(BaseInputWithIMUData):
    external_id: str
    frames: List[Frame]
    metadata: AllowedMetaData = dict()

    @property
    def resources(self) -> Mapping[str, PointCloud]:
        mappings = [frame.resources for frame in self.frames]
        superset = {}
        for mapping in mappings:
            superset = {**superset, **mapping}
        return superset



"""Client for communicating with the Kognic platform."""
import logging
from typing import List, Optional

import kognic.io.model.calibration as CalibrationModel
from kognic.io.resources.abstract import IOResource

log = logging.getLogger(__name__)


class CalibrationResource(IOResource):

    def get_calibration(self, id: str) -> CalibrationModel.SensorCalibrationEntry:  # pylint: disable=redefined-builtin
        """
        Queries the Input API for a specific calibration for the given id

        :param id: The id of the calibration to get
        :return CalibrationModel.SensorCalibrationEntry: The calibration entry
        """

        json_resp = self._client.get('v1/calibrations', params={"id": id})
        return CalibrationModel.SensorCalibrationEntry.from_json(js=json_resp)

    def get_calibrations(self, external_id: Optional[str] = None) -> List[CalibrationModel.SensorCalibrationEntry]:
        """
        Queries the Input API for a list of calibrations available, filtered on external_id
        if provided.

        :param external_id: The external_id to filter on.
        :return List[CalibrationModel.SensorCalibrationEntry]: A list of calibration entries
        """

        json_resp = self._client.get('v1/calibrations', params={"externalId": external_id})

        return [CalibrationModel.SensorCalibrationEntry.from_json(js) for js in json_resp] if isinstance(json_resp, list) \
            else [CalibrationModel.SensorCalibrationEntry.from_json(json_resp)]

    def create_calibration(self, sensor_calibration: CalibrationModel.SensorCalibration) -> CalibrationModel.SensorCalibrationEntry:
        """
        Creates a new calibration, given the SensorCalibration
        :param sensor_calibration: A SensorCalibration instance containing everything to create a calibration.
        :return SensorCalibrationEntry: Class containing the calibration id, external id and time of creation.
        """
        json_resp = self._client.post("v1/calibrations", json=sensor_calibration.to_dict())
        return CalibrationModel.SensorCalibrationEntry.from_json(json_resp)



from typing import Optional, List

from kognic.io.model import IMUData
from kognic.io.model.calibration.common import BaseSerializer


class ValidatedIMUData(BaseSerializer):
    resource_id: str
    signed_url: Optional[str]
    imu_data: List[IMUData]

    def serialize_imu_data(self):
        return [imud.to_dict() for imud in self.imu_data]


class ValidateIMUDataRequest(BaseSerializer):
    imu_data: List[IMUData]
    internal_id: str



from typing import Optional

from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.resources.abstract import IOResource
from kognic.openlabel.models.models import OpenLabelAnnotation


class CreatePreannotationRequest(BaseSerializer):
    scene_uuid: str
    pre_annotation: OpenLabelAnnotation


class PreAnnotationResource(IOResource):
    """
    Resource exposing Kognic PreAnnotations
    """

    def create(self, scene_uuid: str, pre_annotation: OpenLabelAnnotation, dryrun: bool) -> Optional[dict]:
        """
        Upload pre-annotation to a previously created scene.
        This is not possible to do if the scene already have inputs created for it

        :param scene_uuid: the uuid for the scene. Will be the input uuid when input is created
        :param pre_annotation: PreAnnotation on the OpenLabel format
        :param dryrun: If True the files/metadata will be validated but no pre-annotation will be created
        """
        pre_anno_request = CreatePreannotationRequest(scene_uuid=scene_uuid, pre_annotation=pre_annotation)
        return self._client.post("v1/pre-annotations", json=pre_anno_request.to_dict(), dryrun=dryrun, discard_response=True)



from __future__ import annotations

from enum import EnumMeta, Enum
from typing import Set, Dict


class FeatureFlags:
    """
    Holds a set of flags to control optional features of the input creation process.
    """

    class PointCloudFeatures(Enum):
        MOTION_COMPENSATION = "shouldMotionCompensate"

    def __init__(self, *flags):
        self.flags = set(flags)

    @staticmethod
    def defaults() -> FeatureFlags:
        return FeatureFlags(FeatureFlags.PointCloudFeatures.MOTION_COMPENSATION)

    @staticmethod
    def _all() -> Set[Enum]:
        all_flags = set()
        for flag_set in FeatureFlags.__dict__.values():
            if isinstance(flag_set, EnumMeta):
                all_flags.update([flag_set[m] for m in flag_set.__members__])
        return all_flags

    def to_dict(self) -> Dict[str, bool]:
        return dict([(a.value, a in self.flags) for a in FeatureFlags._all()])

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join([type(f).__name__ + '.' + f.name for f in self.flags])})"



import logging
from typing import List, Optional

import kognic.io.model.input as InputModel
from kognic.io.model.input.feature_flags import FeatureFlags
from kognic.io.resources.abstract import CreateableIOResource

log = logging.getLogger(__name__)


class LidarsAndCamerasSequence(CreateableIOResource):

    path = 'lidars-and-cameras-sequence'

    def create(
        self,
        lidars_and_cameras_sequence: InputModel.LidarsAndCamerasSequence,
        project: Optional[str] = None,
        batch: Optional[str] = None,
        annotation_types: Optional[List[str]] = None,
        dryrun: bool = False,
        feature_flags: Optional[FeatureFlags] = None
    ) -> Optional[InputModel.CreateInputResponse]:
        """
        Upload files and create an input of type ``LidarsAndCamerasSequence``.

        :param lidars_and_cameras_sequence: class containing 2D and 3D resources that constitute the input
        :param project: project to add input to
        :param batch: batch, defaults to latest open batch
        :param annotation_types: annotation types for which to produce annotations for. Defaults to `None` (corresponds
         to all available annotation types). Passing an empty list will result in the same behaviour as passing `None`.
        :param dryrun: If True the files/metadata will be validated but no input job will be created.
        :param feature_flags Optional set of feature flags to control the input creation process.
        :returns InputJobCreated: Class containing id of the created input job, or `None` if dryrun.
        """
        if not isinstance(lidars_and_cameras_sequence, InputModel.LidarsAndCamerasSequence):
            raise ValueError(f"Cannot create a {type(lidars_and_cameras_sequence)} via this endpoint")

        response = self._post_input_request(
            self.path,
            lidars_and_cameras_sequence.to_dict(),
            project=project,
            batch=batch,
            annotation_types=annotation_types,
            imu_data=lidars_and_cameras_sequence.imu_data,
            resources=lidars_and_cameras_sequence.resources,
            dryrun=dryrun,
            feature_flags=feature_flags
        )

        if dryrun:
            return None

        log.info(f"Created inputs for files with uuid={response.input_uuid}")
        return InputModel.CreateInputResponse.from_input_job_response(response)



from typing import Optional

from pydantic import validator

from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.model.input.resources.resource import Resource

camera_sensor_default = "CAM"


class ImageMetadata(BaseSerializer):
    shutter_time_start_ns: int
    shutter_time_end_ns: int


class Image(Resource):
    sensor_name: str = camera_sensor_default
    metadata: Optional[ImageMetadata] = None

    @validator('file_data', pre=True)
    def cannot_be_pointcloud(cls, value):
        if not value.format.is_image:
            raise ValueError(f"Invalid format for image data: {value.format}")
        return value



from kognic.io.model.calibration.common import BaseSerializer, Position, RotationQuaternion
from kognic.io.model.ego.utils import UnixTimestampNs


class IMUData(BaseSerializer):
    # TODO: Check that the following statement is true
    """ Data from the Inertial Measurement Unit, both `position` and `rotation` are with respect to the local
        coordinate system (LCS).
    """
    position: Position
    rotation_quaternion: RotationQuaternion
    timestamp: UnixTimestampNs



from typing import List

from kognic.io.model.base_serializer import BaseSerializer


class FilesToUpload(BaseSerializer):
    """
    Used when retrieving upload urls from input api
    """
    files_to_upload: List[str]



from typing import Optional

from kognic.io.model.calibration.common import BaseSerializer, BaseCalibration


class CameraMatrix(BaseSerializer):
    fx: float
    fy: float
    cx: float
    cy: float


class DistortionCoefficients(BaseSerializer):
    k1: float
    k2: float
    p1: float
    p2: float
    k3: float


class BaseCameraCalibration(BaseCalibration):
    camera_matrix: CameraMatrix
    image_height: int
    image_width: int
    field_of_view: Optional[float] = None



import mimetypes
from abc import ABC
from pathlib import Path
from typing import Optional

from pydantic import root_validator, Field

from kognic.io.model.base_serializer import BaseSerializer
from kognic.io.resources.input.file_data import FileData


class Resource(ABC, BaseSerializer):
    filename: str
    resource_id: Optional[str] = None
    sensor_name: str
    file_data: Optional[FileData] = Field(default=None, exclude=True)
    client_filename: Optional[str]

    class Config(BaseSerializer.Config):
        arbitrary_types_allowed = {'file_data': True}

    @property
    def content_type(self):
        if self.file_data:
            return self.file_data.content_type
        content_type = mimetypes.guess_type(self.filename)[0]
        return content_type if content_type is not None else 'application/octet-stream'

    @root_validator(pre=True)
    def check_source_of_data(cls, values):
        filename = values.get('filename')
        resource_id = values.get('resource_id')
        file_data = values.get('file_data')
        values['client_filename'] = filename
        filename_path = Resource.filename_path(filename, file_data)
        values['filename'] = str(filename_path)
        if resource_id is None:
            values['resource_id'] = Resource.resource_id_from_filename(filename_path)
        return values

    @staticmethod
    def filename_path(filename: str, file_data: FileData):
        return Resource.suffix_for_data(Path(filename).expanduser(), file_data)

    @staticmethod
    def suffix_for_data(path: Path, file_data: FileData) -> Path:
        if file_data is None:
            return path
        return path if path.suffix == file_data.suffix else path.with_name(path.stem + path.suffix + file_data.suffix)

    @staticmethod
    def resource_id_from_filename(filename_path: Path) -> str:
        """
        Protects against cloud storage limitation of one period per resource by replacing any others.
        """
        path = filename_path
        filename = path.name
        if '.' in filename:
            path = path.with_name(filename.replace('.', '_', filename.count('.') - 1))
        return str(path)



from typing import List, Mapping, Union, Optional

from kognic.io.model.ego import EgoVehiclePose
from kognic.io.model.ego.utils import UnixTimestampNs
from kognic.io.model.input.abstract.sequence_frame import SequenceFrame
from kognic.io.model.input.resources import PointCloud, Image, VideoFrame


class Frame(SequenceFrame):
    point_clouds: List[PointCloud]
    images: List[Image] = []
    video_frames: List[VideoFrame] = []
    ego_vehicle_pose: Optional[EgoVehiclePose] = None
    metadata: Mapping[str, Union[int, float, str, bool]] = {}
    unix_timestamp: Optional[UnixTimestampNs] = None

    @property
    def resources(self) -> Mapping[str, Image]:
        # Video is not included as of 2022-10; currently unused and unsupported.
        img_mappings = {i.resource_id: i for i in self.images}
        pc_mappings = {p.resource_id: p for p in self.point_clouds}
        return {**img_mappings, **pc_mappings}



from typing import List, Mapping

from kognic.io.model.input.abstract.base_frame import BaseFrame
from kognic.io.model.input.resources.image import Image


class Frame(BaseFrame):
    images: List[Image]

    @property
    def resources(self) -> Mapping[str, Image]:
        return {i.resource_id: i for i in self.images}



from kognic.io.model.calibration.camera.common import CameraMatrix, DistortionCoefficients, BaseCameraCalibration
from kognic.io.model.calibration.camera.fisheye_calibration import FisheyeCalibration
from kognic.io.model.calibration.camera.kannala_calibration import KannalaCalibration
from kognic.io.model.calibration.camera.pinhole_calibration import PinholeCalibration
from kognic.io.model.calibration.camera.principal_point_distortion_calibration import \
    PrincipalPointDistortionCalibration



from typing import Optional, List, Dict

from kognic.io.model.base_serializer import BaseSerializer


class CameraSettings(BaseSerializer):
    width: int
    height: int


class SensorSpecification(BaseSerializer):
    sensor_to_pretty_name: Optional[Dict[str, str]] = None
    sensor_order: Optional[List[str]] = None



from typing import Optional, Generator

from kognic.io.model.annotation.client_annotation import Annotation, PartialAnnotation
from kognic.io.resources.abstract import IOResource


class AnnotationResource(IOResource):

    def get_project_annotations(self, project: str, annotation_type: str, batch: Optional[str] = None) -> Generator[Annotation, None, None]:
        url = f"v1/annotations/projects/{project}/"
        if batch:
            url += f"batch/{batch}/"

        url += f"annotation-type/{annotation_type}/search"

        for js in self._paginate_get(url):
            partial_annotation = PartialAnnotation.from_json(js)
            content = self._file_client.get_json(partial_annotation.uri)
            yield partial_annotation.to_annotation(content)

    def get_annotation(self, input_uuid: str, annotation_type: str) -> Annotation:
        json_resp = self._client.get(f"v1/annotations/inputs/{input_uuid}/annotation-type/{annotation_type}")
        annotation = Annotation.from_json(json_resp)
        return annotation
